@@ -32,37 +32,25 @@
 #include <glib.h>
 
 #include <bluetooth/bluetooth.h>
-#include <bluetooth/sdp.h>
-#include <bluetooth/sdp_lib.h>
+#include <bluetooth/uuid.h>
 
 #include "adapter.h"
 #include "device.h"
 #include "log.h"
 #include "gdbus.h"
 #include "error.h"
-#include "glib-helper.h"
 #include "dbus-common.h"
 #include "btio.h"
 #include "storage.h"
 
 #include "att.h"
 #include "gattrib.h"
+#include "attio.h"
 #include "gatt.h"
 #include "client.h"
 
 #define CHAR_INTERFACE "org.bluez.Characteristic"
 
-struct gatt_service {
-	struct btd_device *dev;
-	bdaddr_t sba;
-	bdaddr_t dba;
-	char *path;
-	GSList *primary;
-	GAttrib *attrib;
-	int psm;
-	gboolean listen;
-};
-
 struct format {
 	guint8 format;
 	guint8 exponent;
@@ -71,21 +59,32 @@ struct format {
 	guint16 desc;
 } __attribute__ ((packed));
 
-struct primary {
-	struct gatt_service *gatt;
-	struct att_primary *att;
+struct query {
+	DBusMessage *msg;
+	GSList *list;
+};
+
+struct gatt_service {
+	struct btd_device *dev;
+	struct gatt_primary *prim;
+	DBusConnection *conn;
+	GAttrib *attrib;
+	guint attioid;
+	int psm;
 	char *path;
 	GSList *chars;
+	GSList *offline_chars;
 	GSList *watchers;
+	struct query *query;
 };
 
 struct characteristic {
-	struct primary *prim;
+	struct gatt_service *gatt;
 	char *path;
 	uint16_t handle;
 	uint16_t end;
 	uint8_t perm;
-	uuid_t type;
+	char type[MAX_LEN_UUID_STR + 1];
 	char *name;
 	char *desc;
 	struct format *format;
@@ -94,7 +93,7 @@ struct characteristic {
 };
 
 struct query_data {
-	struct primary *prim;
+	struct gatt_service *gatt;
 	struct characteristic *chr;
 	uint16_t handle;
 };
@@ -103,13 +102,11 @@ struct watcher {
 	guint id;
 	char *name;
 	char *path;
-	struct primary *prim;
+	struct gatt_service *gatt;
 };
 
 static GSList *gatt_services = NULL;
 
-static DBusConnection *connection;
-
 static void characteristic_free(void *user_data)
 {
 	struct characteristic *chr = user_data;
@@ -131,40 +128,42 @@ static void watcher_free(void *user_data)
 	g_free(watcher);
 }
 
-static void primary_free(void *user_data)
+static void gatt_service_free(struct gatt_service *gatt)
 {
-	struct primary *prim = user_data;
-	GSList *l;
-
-	for (l = prim->watchers; l; l = l->next) {
-		struct watcher *watcher = l->data;
-		g_dbus_remove_watch(connection, watcher->id);
-	}
-
-	g_slist_foreach(prim->chars, (GFunc) characteristic_free, NULL);
-	g_slist_free(prim->chars);
-	g_free(prim->path);
-	g_free(prim);
-}
-
-static void gatt_service_free(void *user_data)
-{
-	struct gatt_service *gatt = user_data;
-
-	g_slist_foreach(gatt->primary, (GFunc) primary_free, NULL);
-	g_slist_free(gatt->primary);
-	g_attrib_unref(gatt->attrib);
+	g_slist_free_full(gatt->watchers, watcher_free);
+	g_slist_free_full(gatt->chars, characteristic_free);
+	g_slist_free(gatt->offline_chars);
 	g_free(gatt->path);
 	btd_device_unref(gatt->dev);
+	dbus_connection_unref(gatt->conn);
 	g_free(gatt);
 }
 
-static int gatt_dev_cmp(gconstpointer a, gconstpointer b)
+static void remove_attio(struct gatt_service *gatt)
 {
-	const struct gatt_service *gatt = a;
-	const struct btd_device *dev = b;
+	if (gatt->offline_chars || gatt->watchers || gatt->query)
+		return;
 
-	return gatt->dev != dev;
+	if (gatt->attioid) {
+		btd_device_remove_attio_callback(gatt->dev, gatt->attioid);
+		gatt->attioid = 0;
+	}
+
+	if (gatt->attrib) {
+		g_attrib_unref(gatt->attrib);
+		gatt->attrib = NULL;
+	}
+}
+
+static void gatt_get_address(struct gatt_service *gatt, bdaddr_t *sba,
+					bdaddr_t *dba, uint8_t *bdaddr_type)
+{
+	struct btd_device *device = gatt->dev;
+	struct btd_adapter *adapter;
+
+	adapter = device_get_adapter(device);
+	adapter_get_address(adapter, sba);
+	device_get_address(device, dba, bdaddr_type);
 }
 
 static int characteristic_handle_cmp(gconstpointer a, gconstpointer b)
@@ -199,7 +198,7 @@ static void append_char_dict(DBusMessageIter *iter, struct characteristic *chr)
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	uuid = bt_uuid2string(&chr->type);
+	uuid = g_strdup(chr->type);
 	dict_append_entry(&dict, "UUID", DBUS_TYPE_STRING, &uuid);
 	g_free(uuid);
 
@@ -222,14 +221,13 @@ static void append_char_dict(DBusMessageIter *iter, struct characteristic *chr)
 static void watcher_exit(DBusConnection *conn, void *user_data)
 {
 	struct watcher *watcher = user_data;
-	struct primary *prim = watcher->prim;
-	struct gatt_service *gatt = prim->gatt;
+	struct gatt_service *gatt = watcher->gatt;
 
-	DBG("%s watcher %s exited", prim->path, watcher->name);
+	DBG("%s watcher %s exited", gatt->path, watcher->name);
 
-	prim->watchers = g_slist_remove(prim->watchers, watcher);
-
-	g_attrib_unref(gatt->attrib);
+	gatt->watchers = g_slist_remove(gatt->watchers, watcher);
+	g_dbus_remove_watch(gatt->conn, watcher->id);
+	remove_attio(gatt);
 }
 
 static int characteristic_set_value(struct characteristic *chr,
@@ -249,6 +247,7 @@ static void update_watchers(gpointer data, gpointer user_data)
 {
 	struct watcher *w = data;
 	struct characteristic *chr = user_data;
+	DBusConnection *conn = w->gatt->conn;
 	DBusMessage *msg;
 
 	msg = dbus_message_new_method_call(w->name, w->path,
@@ -261,7 +260,7 @@ static void update_watchers(gpointer data, gpointer user_data)
 			&chr->value, chr->vlen, DBUS_TYPE_INVALID);
 
 	dbus_message_set_no_reply(msg, TRUE);
-	g_dbus_send_message(connection, msg);
+	g_dbus_send_message(conn, msg);
 }
 
 static void events_handler(const uint8_t *pdu, uint16_t len,
@@ -269,24 +268,26 @@ static void events_handler(const uint8_t *pdu, uint16_t len,
 {
 	struct gatt_service *gatt = user_data;
 	struct characteristic *chr;
-	struct primary *prim;
-	GSList *lprim, *lchr;
+	GSList *l;
 	uint8_t opdu[ATT_MAX_MTU];
-	guint handle = att_get_u16(&pdu[1]);
+	guint handle;
 	uint16_t olen;
 
-	for (lprim = gatt->primary, prim = NULL, chr = NULL; lprim;
-						lprim = lprim->next) {
-		prim = lprim->data;
-
-		lchr = g_slist_find_custom(prim->chars,
-			GUINT_TO_POINTER(handle), characteristic_handle_cmp);
-		if (lchr) {
-			chr = lchr->data;
-			break;
-		}
+	if (len < 3) {
+		DBG("Malformed notification/indication packet (opcode 0x%02x)",
+									pdu[0]);
+		return;
 	}
 
+	handle = att_get_u16(&pdu[1]);
+
+	l = g_slist_find_custom(gatt->chars, GUINT_TO_POINTER(handle),
+						characteristic_handle_cmp);
+	if (!l)
+		return;
+
+	chr = l->data;
+
 	if (chr == NULL) {
 		DBG("Attribute handle 0x%02x not found", handle);
 		return;
@@ -295,130 +296,113 @@ static void events_handler(const uint8_t *pdu, uint16_t len,
 	switch (pdu[0]) {
 	case ATT_OP_HANDLE_IND:
 		olen = enc_confirmation(opdu, sizeof(opdu));
-		g_attrib_send(gatt->attrib, opdu[0], opdu, olen,
+		g_attrib_send(gatt->attrib, 0, opdu[0], opdu, olen,
 						NULL, NULL, NULL);
 	case ATT_OP_HANDLE_NOTIFY:
 		if (characteristic_set_value(chr, &pdu[3], len - 3) < 0)
-			DBG("Can't change Characteristic %0x02x", handle);
+			DBG("Can't change Characteristic 0x%02x", handle);
 
-		g_slist_foreach(prim->watchers, update_watchers, chr);
+		g_slist_foreach(gatt->watchers, update_watchers, chr);
 		break;
 	}
 }
 
-static void attrib_destroy(gpointer user_data)
+static void offline_char_written(gpointer user_data)
 {
-	struct gatt_service *gatt = user_data;
+	struct characteristic *chr = user_data;
+	struct gatt_service *gatt = chr->gatt;
 
-	gatt->attrib = NULL;
+	gatt->offline_chars = g_slist_remove(gatt->offline_chars, chr);
+
+	remove_attio(gatt);
 }
 
-static void attrib_disconnect(gpointer user_data)
+static void offline_char_write(gpointer data, gpointer user_data)
 {
-	struct gatt_service *gatt = user_data;
+	struct characteristic *chr = data;
+	GAttrib *attrib = user_data;
 
-	/* Remote initiated disconnection only */
-	g_attrib_unref(gatt->attrib);
+	gatt_write_cmd(attrib, chr->handle, chr->value, chr->vlen,
+						offline_char_written, chr);
 }
 
-static void connect_cb(GIOChannel *chan, GError *gerr, gpointer user_data)
+static void char_discovered_cb(GSList *characteristics, guint8 status,
+							gpointer user_data);
+
+static void attio_connected(GAttrib *attrib, gpointer user_data)
 {
 	struct gatt_service *gatt = user_data;
 
-	if (gerr) {
-		error("%s", gerr->message);
-		goto fail;
-	}
+	gatt->attrib = g_attrib_ref(attrib);
 
-	if (gatt->attrib == NULL)
-		return;
-
-	/* Listen mode: used for notification and indication */
-	if (gatt->listen == TRUE) {
-		g_attrib_register(gatt->attrib,
-					ATT_OP_HANDLE_NOTIFY,
+	g_attrib_register(gatt->attrib, ATT_OP_HANDLE_NOTIFY,
 					events_handler, gatt, NULL);
-		g_attrib_register(gatt->attrib,
-					ATT_OP_HANDLE_IND,
+	g_attrib_register(gatt->attrib, ATT_OP_HANDLE_IND,
 					events_handler, gatt, NULL);
-		return;
-	}
 
-	return;
-fail:
-	g_attrib_unref(gatt->attrib);
+	g_slist_foreach(gatt->offline_chars, offline_char_write, attrib);
+
+	if (gatt->query) {
+		struct gatt_primary *prim = gatt->prim;
+		struct query_data *qchr;
+
+		qchr = g_slist_nth_data(gatt->query->list, 0);
+		gatt_discover_char(gatt->attrib, prim->range.start,
+						prim->range.end, NULL,
+						char_discovered_cb, qchr);
+	}
 }
 
-static int l2cap_connect(struct gatt_service *gatt, GError **gerr,
-								gboolean listen)
+static void attio_disconnected(gpointer user_data)
 {
-	GIOChannel *io;
+	struct gatt_service *gatt = user_data;
 
-	if (gatt->attrib != NULL) {
-		gatt->attrib = g_attrib_ref(gatt->attrib);
-		return 0;
+	if (gatt->query && gatt->query->msg) {
+		DBusMessage *reply;
+
+		reply = btd_error_failed(gatt->query->msg,
+					"ATT IO channel was disconnected");
+		g_dbus_send_message(gatt->conn, reply);
+		dbus_message_unref(gatt->query->msg);
 	}
 
-	/*
-	 * FIXME: If the service doesn't support Client Characteristic
-	 * Configuration it is necessary to poll the server from time
-	 * to time checking for modifications.
-	 */
-	if (gatt->psm < 0)
-		io = bt_io_connect(BT_IO_L2CAP, connect_cb, gatt, NULL, gerr,
-			BT_IO_OPT_SOURCE_BDADDR, &gatt->sba,
-			BT_IO_OPT_DEST_BDADDR, &gatt->dba,
-			BT_IO_OPT_CID, GATT_CID,
-			BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
-			BT_IO_OPT_INVALID);
-	else
-		io = bt_io_connect(BT_IO_L2CAP, connect_cb, gatt, NULL, gerr,
-			BT_IO_OPT_SOURCE_BDADDR, &gatt->sba,
-			BT_IO_OPT_DEST_BDADDR, &gatt->dba,
-			BT_IO_OPT_PSM, gatt->psm,
-			BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_LOW,
-			BT_IO_OPT_INVALID);
-	if (!io)
-		return -1;
+	if (gatt->query) {
+		g_slist_free_full(gatt->query->list, g_free);
+		gatt->query = NULL;
+	}
 
-	gatt->attrib = g_attrib_new(io);
-	g_io_channel_unref(io);
-	gatt->listen = listen;
-
-	g_attrib_set_destroy_function(gatt->attrib, attrib_destroy, gatt);
-	g_attrib_set_disconnect_function(gatt->attrib, attrib_disconnect,
-									gatt);
-
-	return 0;
+	if (gatt->attrib) {
+		g_attrib_unref(gatt->attrib);
+		gatt->attrib = NULL;
+	}
 }
 
 static DBusMessage *register_watcher(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	const char *sender = dbus_message_get_sender(msg);
-	struct primary *prim = data;
+	struct gatt_service *gatt = data;
 	struct watcher *watcher;
-	GError *gerr = NULL;
 	char *path;
 
 	if (!dbus_message_get_args(msg, NULL, DBUS_TYPE_OBJECT_PATH, &path,
 							DBUS_TYPE_INVALID))
 		return btd_error_invalid_args(msg);
 
-	if (l2cap_connect(prim->gatt, &gerr, TRUE) < 0) {
-		DBusMessage *reply = btd_error_failed(msg, gerr->message);
-		g_error_free(gerr);
-		return reply;
-	}
-
 	watcher = g_new0(struct watcher, 1);
 	watcher->name = g_strdup(sender);
-	watcher->prim = prim;
+	watcher->gatt = gatt;
 	watcher->path = g_strdup(path);
 	watcher->id = g_dbus_add_disconnect_watch(conn, sender, watcher_exit,
 							watcher, watcher_free);
 
-	prim->watchers = g_slist_append(prim->watchers, watcher);
+	if (gatt->attioid == 0)
+		gatt->attioid = btd_device_add_attio_callback(gatt->dev,
+							attio_connected,
+							attio_disconnected,
+							gatt);
+
+	gatt->watchers = g_slist_append(gatt->watchers, watcher);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -427,7 +411,7 @@ static DBusMessage *unregister_watcher(DBusConnection *conn,
 						DBusMessage *msg, void *data)
 {
 	const char *sender = dbus_message_get_sender(msg);
-	struct primary *prim = data;
+	struct gatt_service *gatt = data;
 	struct watcher *watcher, *match;
 	GSList *l;
 	char *path;
@@ -439,15 +423,15 @@ static DBusMessage *unregister_watcher(DBusConnection *conn,
 	match = g_new0(struct watcher, 1);
 	match->name = g_strdup(sender);
 	match->path = g_strdup(path);
-	l = g_slist_find_custom(prim->watchers, match, watcher_cmp);
+	l = g_slist_find_custom(gatt->watchers, match, watcher_cmp);
 	watcher_free(match);
 	if (!l)
 		return btd_error_not_authorized(msg);
 
 	watcher = l->data;
+	gatt->watchers = g_slist_remove(gatt->watchers, watcher);
 	g_dbus_remove_watch(conn, watcher->id);
-	prim->watchers = g_slist_remove(prim->watchers, watcher);
-	watcher_free(watcher);
+	remove_attio(gatt);
 
 	return dbus_message_new_method_return(msg);
 }
@@ -455,9 +439,8 @@ static DBusMessage *unregister_watcher(DBusConnection *conn,
 static DBusMessage *set_value(DBusConnection *conn, DBusMessage *msg,
 			DBusMessageIter *iter, struct characteristic *chr)
 {
-	struct gatt_service *gatt = chr->prim->gatt;
+	struct gatt_service *gatt = chr->gatt;
 	DBusMessageIter sub;
-	GError *gerr = NULL;
 	uint8_t *value;
 	int len;
 
@@ -469,16 +452,20 @@ static DBusMessage *set_value(DBusConnection *conn, DBusMessage *msg,
 
 	dbus_message_iter_get_fixed_array(&sub, &value, &len);
 
-	if (l2cap_connect(gatt, &gerr, FALSE) < 0) {
-		DBusMessage *reply = btd_error_failed(msg, gerr->message);
-		g_error_free(gerr);
-		return reply;
-	}
-
-	gatt_write_cmd(gatt->attrib, chr->handle, value, len, NULL, NULL);
-
 	characteristic_set_value(chr, value, len);
 
+	if (gatt->attioid == 0)
+		gatt->attioid = btd_device_add_attio_callback(gatt->dev,
+							attio_connected,
+							attio_disconnected,
+							gatt);
+
+	if (gatt->attrib)
+		gatt_write_cmd(gatt->attrib, chr->handle, value, len,
+								NULL, NULL);
+	else
+		gatt->offline_chars = g_slist_append(gatt->offline_chars, chr);
+
 	return dbus_message_new_method_return(msg);
 }
 
@@ -528,10 +515,13 @@ static DBusMessage *set_property(DBusConnection *conn,
 	return btd_error_invalid_args(msg);
 }
 
-static GDBusMethodTable char_methods[] = {
-	{ "GetProperties",	"",	"a{sv}", get_properties },
-	{ "SetProperty",	"sv",	"",	set_property,
-						G_DBUS_METHOD_FLAG_ASYNC},
+static const GDBusMethodTable char_methods[] = {
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			get_properties) },
+	{ GDBUS_METHOD("SetProperty",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" }), NULL,
+			set_property) },
 	{ }
 };
 
@@ -544,19 +534,12 @@ static char *characteristic_list_to_string(GSList *chars)
 
 	for (l = chars; l; l = l->next) {
 		struct characteristic *chr = l->data;
-		uuid_t *uuid128;
 		char chr_str[64];
-		char uuidstr[MAX_LEN_UUID_STR];
 
 		memset(chr_str, 0, sizeof(chr_str));
 
-		uuid128 = sdp_uuid_to_uuid128(&chr->type);
-		sdp_uuid2strn(uuid128, uuidstr, MAX_LEN_UUID_STR);
-
-		bt_free(uuid128);
-
 		snprintf(chr_str, sizeof(chr_str), "%04X#%02X#%04X#%s ",
-				chr->handle, chr->perm, chr->end, uuidstr);
+				chr->handle, chr->perm, chr->end, chr->type);
 
 		characteristics = g_string_append(characteristics, chr_str);
 	}
@@ -564,34 +547,36 @@ static char *characteristic_list_to_string(GSList *chars)
 	return g_string_free(characteristics, FALSE);
 }
 
-static void store_characteristics(struct gatt_service *gatt,
-							struct primary *prim)
+static void store_characteristics(const bdaddr_t *sba, const bdaddr_t *dba,
+					uint8_t bdaddr_type, uint16_t start,
+								GSList *chars)
 {
 	char *characteristics;
-	struct att_primary *att = prim->att;
 
-	characteristics = characteristic_list_to_string(prim->chars);
+	characteristics = characteristic_list_to_string(chars);
 
-	write_device_characteristics(&gatt->sba, &gatt->dba, att->start,
+	write_device_characteristics(sba, dba, bdaddr_type, start,
 							characteristics);
 
 	g_free(characteristics);
 }
 
-static void register_characteristics(struct primary *prim)
+static void register_characteristic(gpointer data, gpointer user_data)
 {
-	GSList *lc;
+	struct characteristic *chr = data;
+	DBusConnection *conn = chr->gatt->conn;
+	const char *gatt_path = user_data;
 
-	for (lc = prim->chars; lc; lc = lc->next) {
-		struct characteristic *chr = lc->data;
-		g_dbus_register_interface(connection, chr->path,
-				CHAR_INTERFACE, char_methods,
-				NULL, NULL, chr, NULL);
-		DBG("Registered: %s", chr->path);
-	}
+	chr->path = g_strdup_printf("%s/characteristic%04x", gatt_path,
+								chr->handle);
+
+	g_dbus_register_interface(conn, chr->path, CHAR_INTERFACE,
+					char_methods, NULL, NULL, chr, NULL);
+
+	DBG("Registered: %s", chr->path);
 }
 
-static GSList *string_to_characteristic_list(struct primary *prim,
+static GSList *string_to_characteristic_list(struct gatt_service *gatt,
 							const char *str)
 {
 	GSList *l = NULL;
@@ -607,24 +592,18 @@ static GSList *string_to_characteristic_list(struct primary *prim,
 
 	for (i = 0; chars[i]; i++) {
 		struct characteristic *chr;
-		char uuidstr[MAX_LEN_UUID_STR + 1];
 		int ret;
 
 		chr = g_new0(struct characteristic, 1);
 
 		ret = sscanf(chars[i], "%04hX#%02hhX#%04hX#%s", &chr->handle,
-				&chr->perm, &chr->end, uuidstr);
+				&chr->perm, &chr->end, chr->type);
 		if (ret < 4) {
 			g_free(chr);
 			continue;
 		}
 
-		chr->prim = prim;
-		chr->path = g_strdup_printf("%s/characteristic%04x",
-						prim->path, chr->handle);
-
-		bt_string2uuid(&chr->type, uuidstr);
-
+		chr->gatt = gatt;
 		l = g_slist_append(l, chr);
 	}
 
@@ -633,79 +612,115 @@ static GSList *string_to_characteristic_list(struct primary *prim,
 	return l;
 }
 
-static void load_characteristics(gpointer data, gpointer user_data)
+static GSList *load_characteristics(struct gatt_service *gatt, uint16_t start)
 {
-	struct primary *prim = data;
-	struct att_primary *att = prim->att;
-	struct gatt_service *gatt = user_data;
 	GSList *chrs_list;
+	bdaddr_t sba, dba;
+	uint8_t bdaddr_type;
 	char *str;
 
-	if (prim->chars) {
-		DBG("Characteristics already loaded");
-		return;
-	}
+	gatt_get_address(gatt, &sba, &dba, &bdaddr_type);
 
-	str = read_device_characteristics(&gatt->sba, &gatt->dba, att->start);
+	str = read_device_characteristics(&sba, &dba, bdaddr_type, start);
 	if (str == NULL)
-		return;
+		return NULL;
 
-	chrs_list = string_to_characteristic_list(prim, str);
+	chrs_list = string_to_characteristic_list(gatt, str);
 
 	free(str);
 
-	if (chrs_list == NULL)
-		return;
-
-	prim->chars = chrs_list;
-	register_characteristics(prim);
-
-	return;
+	return chrs_list;
 }
 
 static void store_attribute(struct gatt_service *gatt, uint16_t handle,
 				uint16_t type, uint8_t *value, gsize len)
 {
-	uuid_t uuid;
-	char *str, *uuidstr, *tmp;
+	struct btd_device *device = gatt->dev;
+	bdaddr_t sba, dba;
+	uint8_t bdaddr_type;
+	bt_uuid_t uuid;
+	char *str, *tmp;
 	guint i;
 
 	str = g_malloc0(MAX_LEN_UUID_STR + len * 2 + 1);
 
-	sdp_uuid16_create(&uuid, type);
-	uuidstr = bt_uuid2string(&uuid);
-	strcpy(str, uuidstr);
-	g_free(uuidstr);
+	bt_uuid16_create(&uuid, type);
+	bt_uuid_to_string(&uuid, str, MAX_LEN_UUID_STR);
 
 	str[MAX_LEN_UUID_STR - 1] = '#';
 
 	for (i = 0, tmp = str + MAX_LEN_UUID_STR; i < len; i++, tmp += 2)
 		sprintf(tmp, "%02X", value[i]);
 
-	write_device_attribute(&gatt->sba, &gatt->dba, handle, str);
+	gatt_get_address(gatt, &sba, &dba, NULL);
+
+	bdaddr_type = device_get_addr_type(device);
+
+	write_device_attribute(&sba, &dba, bdaddr_type, handle, str);
+
 	g_free(str);
 }
 
+static void query_list_append(struct gatt_service *gatt, struct query_data *data)
+{
+	struct query *query = gatt->query;
+
+	query->list = g_slist_append(query->list, data);
+}
+
+static void query_list_remove(struct gatt_service *gatt, struct query_data *data)
+{
+	struct query *query = gatt->query;
+
+	query->list = g_slist_remove(query->list, data);
+	if (query->list != NULL)
+		return;
+
+	g_free(query);
+	gatt->query = NULL;
+
+	remove_attio(gatt);
+}
+
 static void update_char_desc(guint8 status, const guint8 *pdu, guint16 len,
 							gpointer user_data)
 {
 	struct query_data *current = user_data;
-	struct gatt_service *gatt = current->prim->gatt;
+	struct gatt_service *gatt = current->gatt;
 	struct characteristic *chr = current->chr;
 
-	if (status != 0)
-		goto done;
+	if (status == 0) {
 
-	g_free(chr->desc);
+		g_free(chr->desc);
 
-	chr->desc = g_malloc(len);
-	memcpy(chr->desc, pdu + 1, len - 1);
-	chr->desc[len - 1] = '\0';
+		chr->desc = g_malloc(len);
+		memcpy(chr->desc, pdu + 1, len - 1);
+		chr->desc[len - 1] = '\0';
 
-	store_attribute(gatt, current->handle, GATT_CHARAC_USER_DESC_UUID,
-						(void *) chr->desc, len);
-done:
-	g_attrib_unref(gatt->attrib);
+		store_attribute(gatt, current->handle,
+				GATT_CHARAC_USER_DESC_UUID,
+				(void *) chr->desc, len);
+	} else if (status == ATT_ECODE_INSUFF_ENC) {
+		GIOChannel *io = g_attrib_get_channel(gatt->attrib);
+		BtIOSecLevel level = BT_IO_SEC_HIGH;
+
+		bt_io_get(io, BT_IO_L2CAP, NULL,
+				BT_IO_OPT_SEC_LEVEL, &level,
+				BT_IO_OPT_INVALID);
+
+		if (level < BT_IO_SEC_HIGH)
+			level++;
+
+		if (bt_io_set(io, BT_IO_L2CAP, NULL,
+				BT_IO_OPT_SEC_LEVEL, level,
+				BT_IO_OPT_INVALID)) {
+			gatt_read_char(gatt->attrib, current->handle, 0,
+					update_char_desc, current);
+			return;
+		}
+	}
+
+	query_list_remove(gatt, current);
 	g_free(current);
 }
 
@@ -713,7 +728,7 @@ static void update_char_format(guint8 status, const guint8 *pdu, guint16 len,
 								gpointer user_data)
 {
 	struct query_data *current = user_data;
-	struct gatt_service *gatt = current->prim->gatt;
+	struct gatt_service *gatt = current->gatt;
 	struct characteristic *chr = current->chr;
 
 	if (status != 0)
@@ -731,7 +746,7 @@ static void update_char_format(guint8 status, const guint8 *pdu, guint16 len,
 				(void *) chr->format, sizeof(*chr->format));
 
 done:
-	g_attrib_unref(gatt->attrib);
+	query_list_remove(gatt, current);
 	g_free(current);
 }
 
@@ -739,30 +754,41 @@ static void update_char_value(guint8 status, const guint8 *pdu,
 					guint16 len, gpointer user_data)
 {
 	struct query_data *current = user_data;
-	struct gatt_service *gatt = current->prim->gatt;
+	struct gatt_service *gatt = current->gatt;
 	struct characteristic *chr = current->chr;
 
 	if (status == 0)
 		characteristic_set_value(chr, pdu + 1, len - 1);
+	else if (status == ATT_ECODE_INSUFF_ENC) {
+		GIOChannel *io = g_attrib_get_channel(gatt->attrib);
 
-	g_attrib_unref(gatt->attrib);
+		if (bt_io_set(io, BT_IO_L2CAP, NULL,
+				BT_IO_OPT_SEC_LEVEL, BT_IO_SEC_HIGH,
+				BT_IO_OPT_INVALID)) {
+			gatt_read_char(gatt->attrib, chr->handle, 0,
+					update_char_value, current);
+			return;
+		}
+	}
+
+	query_list_remove(gatt, current);
 	g_free(current);
 }
 
-static int uuid_desc16_cmp(uuid_t *uuid, guint16 desc)
+static int uuid_desc16_cmp(bt_uuid_t *uuid, guint16 desc)
 {
-	uuid_t u16;
+	bt_uuid_t u16;
 
-	sdp_uuid16_create(&u16, desc);
+	bt_uuid16_create(&u16, desc);
 
-	return sdp_uuid_cmp(uuid, &u16);
+	return bt_uuid_cmp(uuid, &u16);
 }
 
 static void descriptor_cb(guint8 status, const guint8 *pdu, guint16 plen,
 							gpointer user_data)
 {
 	struct query_data *current = user_data;
-	struct gatt_service *gatt = current->prim->gatt;
+	struct gatt_service *gatt = current->gatt;
 	struct att_data_list *list;
 	guint8 format;
 	int i;
@@ -778,14 +804,14 @@ static void descriptor_cb(guint8 status, const guint8 *pdu, guint16 plen,
 
 	for (i = 0; i < list->num; i++) {
 		guint16 handle;
-		uuid_t uuid;
+		bt_uuid_t uuid;
 		uint8_t *info = list->data[i];
 		struct query_data *qfmt;
 
 		handle = att_get_u16(info);
 
 		if (format == 0x01) {
-			sdp_uuid16_create(&uuid, att_get_u16(&info[2]));
+			uuid = att_get_uuid16(&info[2]);
 		} else {
 			/* Currently, only "user description" and "presentation
 			 * format" descriptors are used, and both have 16-bit
@@ -794,17 +820,17 @@ static void descriptor_cb(guint8 status, const guint8 *pdu, guint16 plen,
 			continue;
 		}
 		qfmt = g_new0(struct query_data, 1);
-		qfmt->prim = current->prim;
+		qfmt->gatt = current->gatt;
 		qfmt->chr = current->chr;
 		qfmt->handle = handle;
 
 		if (uuid_desc16_cmp(&uuid, GATT_CHARAC_USER_DESC_UUID) == 0) {
-			gatt->attrib = g_attrib_ref(gatt->attrib);
-			gatt_read_char(gatt->attrib, handle, update_char_desc,
+			query_list_append(gatt, qfmt);
+			gatt_read_char(gatt->attrib, handle, 0, update_char_desc,
 									qfmt);
 		} else if (uuid_desc16_cmp(&uuid, GATT_CHARAC_FMT_UUID) == 0) {
-			gatt->attrib = g_attrib_ref(gatt->attrib);
-			gatt_read_char(gatt->attrib, handle,
+			query_list_append(gatt, qfmt);
+			gatt_read_char(gatt->attrib, handle, 0,
 						update_char_format, qfmt);
 		} else
 			g_free(qfmt);
@@ -812,7 +838,7 @@ static void descriptor_cb(guint8 status, const guint8 *pdu, guint16 plen,
 
 	att_data_list_free(list);
 done:
-	g_attrib_unref(gatt->attrib);
+	query_list_remove(gatt, current);
 	g_free(current);
 }
 
@@ -820,136 +846,163 @@ static void update_all_chars(gpointer data, gpointer user_data)
 {
 	struct query_data *qdesc, *qvalue;
 	struct characteristic *chr = data;
-	struct primary *prim = user_data;
-	struct gatt_service *gatt = prim->gatt;
+	struct gatt_service *gatt = user_data;
 
 	qdesc = g_new0(struct query_data, 1);
-	qdesc->prim = prim;
+	qdesc->gatt = gatt;
 	qdesc->chr = chr;
 
-	gatt->attrib = g_attrib_ref(gatt->attrib);
+	query_list_append(gatt, qdesc);
+
 	gatt_find_info(gatt->attrib, chr->handle + 1, chr->end, descriptor_cb,
 									qdesc);
 
 	qvalue = g_new0(struct query_data, 1);
-	qvalue->prim = prim;
+	qvalue->gatt = gatt;
 	qvalue->chr = chr;
 
-	gatt->attrib = g_attrib_ref(gatt->attrib);
-	gatt_read_char(gatt->attrib, chr->handle, update_char_value, qvalue);
+	query_list_append(gatt, qvalue);
+
+	gatt_read_char(gatt->attrib, chr->handle, 0, update_char_value, qvalue);
 }
 
-static void char_discovered_cb(guint8 status, const guint8 *pdu, guint16 plen,
+static DBusMessage *create_discover_char_reply(DBusMessage *msg, GSList *chars)
+{
+	DBusMessage *reply;
+	DBusMessageIter iter, array_iter;
+	GSList *l;
+
+	reply = dbus_message_new_method_return(msg);
+
+	dbus_message_iter_init_append(reply, &iter);
+
+	dbus_message_iter_open_container(&iter, DBUS_TYPE_ARRAY,
+				DBUS_TYPE_OBJECT_PATH_AS_STRING, &array_iter);
+
+	for (l = chars; l; l = l->next) {
+		struct characteristic *chr = l->data;
+
+		dbus_message_iter_append_basic(&array_iter,
+					DBUS_TYPE_OBJECT_PATH, &chr->path);
+	}
+
+	dbus_message_iter_close_container(&iter, &array_iter);
+
+	return reply;
+}
+
+static void char_discovered_cb(GSList *characteristics, guint8 status,
 							gpointer user_data)
 {
+	DBusMessage *reply;
 	struct query_data *current = user_data;
-	struct primary *prim = current->prim;
-	struct att_primary *att = prim->att;
-	struct gatt_service *gatt = prim->gatt;
-	struct att_data_list *list;
-	uint16_t last, *previous_end = NULL;
-	int i;
-
-	if (status == ATT_ECODE_ATTR_NOT_FOUND)
-		goto done;
+	struct gatt_service *gatt = current->gatt;
+	struct gatt_primary *prim = gatt->prim;
+	uint16_t *previous_end = NULL;
+	GSList *l;
+	bdaddr_t sba, dba;
+	uint8_t bdaddr_type;
 
 	if (status != 0) {
-		DBG("Discover all characteristics failed: %s",
-						att_ecode2str(status));
+		const char *str = att_ecode2str(status);
 
+		DBG("Discover all characteristics failed: %s", str);
+		reply = btd_error_failed(gatt->query->msg, str);
 		goto fail;
 	}
 
-	DBG("Read by Type Response received");
-
-	list = dec_read_by_type_resp(pdu, plen);
-	if (list == NULL)
-		goto fail;
-
-	for (i = 0, last = 0; i < list->num; i++) {
-		uint8_t *decl = list->data[i];
+	for (l = characteristics; l; l = l->next) {
+		struct gatt_char *current_chr = l->data;
 		struct characteristic *chr;
+		guint handle = current_chr->value_handle;
+		GSList *lchr;
+
+		lchr = g_slist_find_custom(gatt->chars,
+			GUINT_TO_POINTER(handle), characteristic_handle_cmp);
+		if (lchr)
+			continue;
 
 		chr = g_new0(struct characteristic, 1);
-		chr->prim = prim;
-		chr->perm = decl[2];
-		chr->handle = att_get_u16(&decl[3]);
-		chr->path = g_strdup_printf("%s/characteristic%04x",
-						prim->path, chr->handle);
-		if (list->len == 7) {
-			sdp_uuid16_create(&chr->type,
-					att_get_u16(&decl[5]));
-		} else
-			sdp_uuid128_create(&chr->type, &decl[5]);
+		chr->gatt = gatt;
+		chr->perm = current_chr->properties;
+		chr->handle = current_chr->value_handle;
+		strncpy(chr->type, current_chr->uuid, sizeof(chr->type));
 
-		if (previous_end) {
-			*previous_end = att_get_u16(decl);
-		}
+		if (previous_end)
+			*previous_end = current_chr->handle;
 
-		last = chr->handle;
 		previous_end = &chr->end;
 
-		prim->chars = g_slist_append(prim->chars, chr);
+		gatt->chars = g_slist_append(gatt->chars, chr);
+		register_characteristic(chr, gatt->path);
 	}
 
 	if (previous_end)
-		*previous_end = att->end;
+		*previous_end = prim->range.end;
 
-	att_data_list_free(list);
+	gatt_get_address(gatt, &sba, &dba, &bdaddr_type);
+	store_characteristics(&sba, &dba, bdaddr_type, prim->range.start,
+								gatt->chars);
 
-	if (last >= att->end)
-		goto done;
+	g_slist_foreach(gatt->chars, update_all_chars, gatt);
 
-	/* Fetch remaining characteristics for the CURRENT primary service */
-	gatt_discover_char(gatt->attrib, last + 1, att->end,
-						char_discovered_cb, current);
-
-	return;
-
-done:
-	store_characteristics(gatt, prim);
-	register_characteristics(prim);
-
-	g_slist_foreach(prim->chars, update_all_chars, prim);
+	reply = create_discover_char_reply(gatt->query->msg, gatt->chars);
 
 fail:
-	g_attrib_unref(gatt->attrib);
+	dbus_message_unref(gatt->query->msg);
+	gatt->query->msg = NULL;
+
+	g_dbus_send_message(gatt->conn, reply);
+	query_list_remove(gatt, current);
 	g_free(current);
 }
 
 static DBusMessage *discover_char(DBusConnection *conn, DBusMessage *msg,
 								void *data)
 {
-	struct primary *prim = data;
-	struct att_primary *att = prim->att;
-	struct gatt_service *gatt = prim->gatt;
+	struct gatt_service *gatt = data;
+	struct query *query;
 	struct query_data *qchr;
-	GError *gerr = NULL;
 
-	if (l2cap_connect(prim->gatt, &gerr, FALSE) < 0) {
-		DBusMessage *reply = btd_error_failed(msg, gerr->message);
-		g_error_free(gerr);
-		return reply;
-	}
+	if (gatt->query)
+		return btd_error_busy(msg);
+
+	query = g_new0(struct query, 1);
 
 	qchr = g_new0(struct query_data, 1);
-	qchr->prim = prim;
+	qchr->gatt = gatt;
 
-	gatt_discover_char(gatt->attrib, att->start, att->end,
+	query->msg = dbus_message_ref(msg);
+
+	if (gatt->attioid == 0) {
+		gatt->attioid = btd_device_add_attio_callback(gatt->dev,
+							attio_connected,
+							attio_disconnected,
+							gatt);
+	} else if (gatt->attrib) {
+		struct gatt_primary *prim = gatt->prim;
+		gatt_discover_char(gatt->attrib, prim->range.start,
+						prim->range.end, NULL,
 						char_discovered_cb, qchr);
+	}
 
-	return dbus_message_new_method_return(msg);
+	gatt->query = query;
+
+	query_list_append(gatt, qchr);
+
+	return NULL;
 }
 
 static DBusMessage *prim_get_properties(DBusConnection *conn, DBusMessage *msg,
 								void *data)
 {
-	struct primary *prim = data;
+	struct gatt_service *gatt = data;
 	DBusMessage *reply;
 	DBusMessageIter iter;
 	DBusMessageIter dict;
 	GSList *l;
 	char **chars;
+	const char *uuid;
 	int i;
 
 	reply = dbus_message_new_method_return(msg);
@@ -963,15 +1016,18 @@ static DBusMessage *prim_get_properties(DBusConnection *conn, DBusMessage *msg,
 			DBUS_TYPE_STRING_AS_STRING DBUS_TYPE_VARIANT_AS_STRING
 			DBUS_DICT_ENTRY_END_CHAR_AS_STRING, &dict);
 
-	chars = g_new0(char *, g_slist_length(prim->chars) + 1);
+	chars = g_new0(char *, g_slist_length(gatt->chars) + 1);
 
-	for (i = 0, l = prim->chars; l; l = l->next, i++) {
+	for (i = 0, l = gatt->chars; l; l = l->next, i++) {
 		struct characteristic *chr = l->data;
 		chars[i] = chr->path;
 	}
 
 	dict_append_array(&dict, "Characteristics", DBUS_TYPE_OBJECT_PATH,
 								&chars, i);
+	uuid = gatt->prim->uuid;
+	dict_append_entry(&dict, "UUID", DBUS_TYPE_STRING, &uuid);
+
 	g_free(chars);
 
 	dbus_message_iter_close_container(&iter, &dict);
@@ -979,108 +1035,110 @@ static DBusMessage *prim_get_properties(DBusConnection *conn, DBusMessage *msg,
 	return reply;
 }
 
-static GDBusMethodTable prim_methods[] = {
-	{ "Discover",		"",	"",		discover_char	},
-	{ "RegisterCharacteristicsWatcher",	"o", "",
-						register_watcher	},
-	{ "UnregisterCharacteristicsWatcher",	"o", "",
-						unregister_watcher	},
-	{ "GetProperties",	"",	"a{sv}",prim_get_properties	},
+static const GDBusMethodTable prim_methods[] = {
+	{ GDBUS_ASYNC_METHOD("DiscoverCharacteristics",
+			NULL, GDBUS_ARGS({ "characteristics", "ao" }),
+			discover_char) },
+	{ GDBUS_METHOD("RegisterCharacteristicsWatcher",
+			GDBUS_ARGS({ "agent", "o" }), NULL,
+			register_watcher) },
+	{ GDBUS_METHOD("UnregisterCharacteristicsWatcher",
+			GDBUS_ARGS({ "agent", "o" }), NULL,
+			unregister_watcher) },
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			prim_get_properties) },
 	{ }
 };
 
-static void register_primaries(struct gatt_service *gatt, GSList *primaries)
+static struct gatt_service *primary_register(DBusConnection *conn,
+						struct btd_device *device,
+						struct gatt_primary *prim,
+						int psm)
 {
-	GSList *l;
-
-	for (l = primaries; l; l = l->next) {
-		struct att_primary *att = l->data;
-		struct primary *prim;
-
-		prim = g_new0(struct primary, 1);
-		prim->att = att;
-		prim->gatt = gatt;
-		prim->path = g_strdup_printf("%s/service%04x", gatt->path,
-								att->start);
-
-		g_dbus_register_interface(connection, prim->path,
-				CHAR_INTERFACE, prim_methods,
-				NULL, NULL, prim, NULL);
-		DBG("Registered: %s", prim->path);
-
-		gatt->primary = g_slist_append(gatt->primary, prim);
-		btd_device_add_service(gatt->dev, prim->path);
-		load_characteristics(prim, gatt);
-	}
-}
-
-int attrib_client_register(struct btd_device *device, int psm)
-{
-	struct btd_adapter *adapter = device_get_adapter(device);
-	const char *path = device_get_path(device);
 	struct gatt_service *gatt;
-	GSList *primaries = btd_device_get_primaries(device);
-	bdaddr_t sba, dba;
+	const char *device_path;
 
-	adapter_get_address(adapter, &sba);
-	device_get_address(device, &dba);
+	device_path = device_get_path(device);
 
 	gatt = g_new0(struct gatt_service, 1);
 	gatt->dev = btd_device_ref(device);
-	gatt->listen = FALSE;
-	gatt->path = g_strdup(path);
-	bacpy(&gatt->sba, &sba);
-	bacpy(&gatt->dba, &dba);
+	gatt->prim = prim;
 	gatt->psm = psm;
+	gatt->conn = dbus_connection_ref(conn);
+	gatt->path = g_strdup_printf("%s/service%04x", device_path,
+								prim->range.start);
 
-	register_primaries(gatt, primaries);
+	g_dbus_register_interface(gatt->conn, gatt->path,
+					CHAR_INTERFACE, prim_methods,
+					NULL, NULL, gatt, NULL);
+	gatt->chars = load_characteristics(gatt, prim->range.start);
+	g_slist_foreach(gatt->chars, register_characteristic, gatt->path);
 
-	gatt_services = g_slist_append(gatt_services, gatt);
-
-	return 0;
+	return gatt;
 }
 
-void attrib_client_unregister(struct btd_device *device)
+GSList *attrib_client_register(DBusConnection *connection,
+					struct btd_device *device, int psm,
+					GAttrib *attrib, GSList *primaries)
 {
-	struct gatt_service *gatt;
-	GSList *l, *lp, *lc;
+	GSList *l, *services;
 
-	l = g_slist_find_custom(gatt_services, device, gatt_dev_cmp);
-	if (!l)
-		return;
+	for (l = primaries, services = NULL; l; l = l->next) {
+		struct gatt_primary *prim = l->data;
+		struct gatt_service *gatt;
 
-	gatt = l->data;
-	gatt_services = g_slist_remove(gatt_services, gatt);
+		gatt = primary_register(connection, device, prim, psm);
+
+		DBG("Registered: %s", gatt->path);
+
+		services = g_slist_append(services, g_strdup(gatt->path));
+		gatt_services = g_slist_append(gatt_services, gatt);
 
-	for (lp = gatt->primary; lp; lp = lp->next) {
-		struct primary *prim = lp->data;
-		for (lc = prim->chars; lc; lc = lc->next) {
-			struct characteristic *chr = lc->data;
-			g_dbus_unregister_interface(connection, chr->path,
-								CHAR_INTERFACE);
-		}
-		g_dbus_unregister_interface(connection, prim->path,
-								CHAR_INTERFACE);
 	}
 
-	gatt_service_free(gatt);
+	return services;
 }
 
-int attrib_client_init(DBusConnection *conn)
+static void primary_unregister(struct gatt_service *gatt)
 {
+	GSList *l;
 
-	connection = dbus_connection_ref(conn);
+	for (l = gatt->chars; l; l = l->next) {
+		struct characteristic *chr = l->data;
+		g_dbus_unregister_interface(gatt->conn, chr->path,
+							CHAR_INTERFACE);
+	}
 
-	/*
-	 * FIXME: if the adapter supports BLE start scanning. Temporary
-	 * solution, this approach doesn't allow to control scanning based
-	 * on the discoverable property.
-	 */
+	g_dbus_unregister_interface(gatt->conn, gatt->path, CHAR_INTERFACE);
 
-	return 0;
+	remove_attio(gatt);
 }
 
-void attrib_client_exit(void)
+static int path_cmp(gconstpointer data, gconstpointer user_data)
 {
-	dbus_connection_unref(connection);
+	const char *path = data;
+	const char *gatt_path = user_data;
+
+	return g_strcmp0(path, gatt_path);
+}
+
+void attrib_client_unregister(GSList *services)
+{
+	GSList *l, *left;
+
+	for (l = gatt_services, left = NULL; l; l = l->next) {
+		struct gatt_service *gatt = l->data;
+
+		if (!g_slist_find_custom(services, gatt->path, path_cmp)) {
+			left = g_slist_append(left, gatt);
+			continue;
+		}
+
+		primary_unregister(gatt);
+		gatt_service_free(gatt);
+	}
+
+	g_slist_free(gatt_services);
+	gatt_services = left;
 }
