@@ -28,6 +28,8 @@
 
 #include <errno.h>
 
+#include <bluetooth/uuid.h>
+
 #include <glib.h>
 #include <gdbus.h>
 
@@ -42,10 +44,8 @@
 #include "transport.h"
 #include "a2dp.h"
 #include "headset.h"
-
-#ifndef DBUS_TYPE_UNIX_FD
-#define DBUS_TYPE_UNIX_FD -1
-#endif
+#include "gateway.h"
+#include "avrcp.h"
 
 #define MEDIA_TRANSPORT_INTERFACE "org.bluez.MediaTransport"
 
@@ -62,11 +62,21 @@ struct media_owner {
 	guint			watch;
 };
 
+struct a2dp_transport {
+	struct avdtp		*session;
+	uint16_t		delay;
+	uint16_t		volume;
+};
+
+struct headset_transport {
+	struct audio_device	*device;
+	unsigned int		nrec_id;
+};
+
 struct media_transport {
 	DBusConnection		*conn;
 	char			*path;		/* Transport object path */
 	struct audio_device	*device;	/* Transport device */
-	struct avdtp		*session;	/* Signalling session (a2dp only) */
 	struct media_endpoint	*endpoint;	/* Transport endpoint */
 	GSList			*owners;	/* Transport owners */
 	uint8_t			*configuration; /* Transport configuration */
@@ -74,8 +84,6 @@ struct media_transport {
 	int			fd;		/* Transport file descriptor */
 	uint16_t		imtu;		/* Transport input mtu */
 	uint16_t		omtu;		/* Transport output mtu */
-	uint16_t		delay;		/* Transport delay (a2dp only) */
-	unsigned int		nrec_id;	/* Transport nrec watch (headset only) */
 	gboolean		read_lock;
 	gboolean		write_lock;
 	gboolean		in_use;
@@ -92,6 +100,8 @@ struct media_transport {
 					struct media_transport *transport,
 					const char *property,
 					DBusMessageIter *value);
+	GDestroyNotify		destroy;
+	void			*data;
 };
 
 void media_transport_destroy(struct media_transport *transport)
@@ -99,7 +109,6 @@ void media_transport_destroy(struct media_transport *transport)
 	char *path;
 
 	path = g_strdup(transport->path);
-
 	g_dbus_unregister_interface(transport->conn, path,
 						MEDIA_TRANSPORT_INTERFACE);
 
@@ -276,26 +285,26 @@ fail:
 static guint resume_a2dp(struct media_transport *transport,
 				struct media_owner *owner)
 {
+	struct a2dp_transport *a2dp = transport->data;
 	struct media_endpoint *endpoint = transport->endpoint;
 	struct audio_device *device = transport->device;
 	struct a2dp_sep *sep = media_endpoint_get_sep(endpoint);
 
-	if (transport->session == NULL) {
-		transport->session = avdtp_get(&device->src, &device->dst);
-		if (transport->session == NULL)
+	if (a2dp->session == NULL) {
+		a2dp->session = avdtp_get(&device->src, &device->dst);
+		if (a2dp->session == NULL)
 			return 0;
 	}
 
 	if (transport->in_use == TRUE)
 		goto done;
 
-	transport->in_use = a2dp_sep_lock(sep, transport->session);
+	transport->in_use = a2dp_sep_lock(sep, a2dp->session);
 	if (transport->in_use == FALSE)
 		return 0;
 
 done:
-	return a2dp_resume(transport->session, sep, a2dp_resume_complete,
-				owner);
+	return a2dp_resume(a2dp->session, sep, a2dp_resume_complete, owner);
 }
 
 static void a2dp_suspend_complete(struct avdtp *session,
@@ -303,6 +312,7 @@ static void a2dp_suspend_complete(struct avdtp *session,
 {
 	struct media_owner *owner = user_data;
 	struct media_transport *transport = owner->transport;
+	struct a2dp_transport *a2dp = transport->data;
 	struct a2dp_sep *sep = media_endpoint_get_sep(transport->endpoint);
 
 	/* Release always succeeds */
@@ -312,7 +322,7 @@ static void a2dp_suspend_complete(struct avdtp *session,
 		media_owner_remove(owner);
 	}
 
-	a2dp_sep_unlock(sep, transport->session);
+	a2dp_sep_unlock(sep, a2dp->session);
 	transport->in_use = FALSE;
 	media_transport_remove(transport, owner);
 }
@@ -320,17 +330,17 @@ static void a2dp_suspend_complete(struct avdtp *session,
 static guint suspend_a2dp(struct media_transport *transport,
 						struct media_owner *owner)
 {
+	struct a2dp_transport *a2dp = transport->data;
 	struct media_endpoint *endpoint = transport->endpoint;
 	struct a2dp_sep *sep = media_endpoint_get_sep(endpoint);
 
 	if (!owner) {
-		a2dp_sep_unlock(sep, transport->session);
+		a2dp_sep_unlock(sep, a2dp->session);
 		transport->in_use = FALSE;
 		return 0;
 	}
 
-	return a2dp_suspend(transport->session, sep, a2dp_suspend_complete,
-				owner);
+	return a2dp_suspend(a2dp->session, sep, a2dp_suspend_complete, owner);
 }
 
 static void cancel_a2dp(struct media_transport *transport, guint id)
@@ -437,6 +447,116 @@ static void cancel_headset(struct media_transport *transport, guint id)
 	headset_cancel_stream(transport->device, id);
 }
 
+static void gateway_resume_complete(struct audio_device *dev, GError *err,
+							void *user_data)
+{
+	struct media_owner *owner = user_data;
+	struct media_request *req = owner->pending;
+	struct media_transport *transport = owner->transport;
+	int fd;
+	uint16_t imtu, omtu;
+	gboolean ret;
+
+	req->id = 0;
+
+	if (dev == NULL)
+		goto fail;
+
+	if (err) {
+		error("Failed to resume gateway: error %s", err->message);
+		goto fail;
+	}
+
+	fd = gateway_get_sco_fd(dev);
+	if (fd < 0)
+		goto fail;
+
+	imtu = 48;
+	omtu = 48;
+
+	media_transport_set_fd(transport, fd, imtu, omtu);
+
+	if (g_strstr_len(owner->accesstype, -1, "r") == NULL)
+		imtu = 0;
+
+	if (g_strstr_len(owner->accesstype, -1, "w") == NULL)
+		omtu = 0;
+
+	ret = g_dbus_send_reply(transport->conn, req->msg,
+						DBUS_TYPE_UNIX_FD, &fd,
+						DBUS_TYPE_UINT16, &imtu,
+						DBUS_TYPE_UINT16, &omtu,
+						DBUS_TYPE_INVALID);
+	if (ret == FALSE)
+		goto fail;
+
+	media_owner_remove(owner);
+
+	return;
+
+fail:
+	media_transport_remove(transport, owner);
+}
+
+static guint resume_gateway(struct media_transport *transport,
+				struct media_owner *owner)
+{
+	struct audio_device *device = transport->device;
+
+	if (transport->in_use == TRUE)
+		goto done;
+
+	transport->in_use = gateway_lock(device, GATEWAY_LOCK_READ |
+						GATEWAY_LOCK_WRITE);
+	if (transport->in_use == FALSE)
+		return 0;
+
+done:
+	return gateway_request_stream(device, gateway_resume_complete,
+					owner);
+}
+
+static gboolean gateway_suspend_complete(gpointer user_data)
+{
+	struct media_owner *owner = user_data;
+	struct media_transport *transport = owner->transport;
+	struct audio_device *device = transport->device;
+
+	/* Release always succeeds */
+	if (owner->pending) {
+		owner->pending->id = 0;
+		media_request_reply(owner->pending, transport->conn, 0);
+		media_owner_remove(owner);
+	}
+
+	gateway_unlock(device, GATEWAY_LOCK_READ | GATEWAY_LOCK_WRITE);
+	transport->in_use = FALSE;
+	media_transport_remove(transport, owner);
+	return FALSE;
+}
+
+static guint suspend_gateway(struct media_transport *transport,
+						struct media_owner *owner)
+{
+	struct audio_device *device = transport->device;
+	static int id = 1;
+
+	if (!owner) {
+		gateway_unlock(device, GATEWAY_LOCK_READ | GATEWAY_LOCK_WRITE);
+		transport->in_use = FALSE;
+		return 0;
+	}
+
+	gateway_suspend_stream(device);
+	g_idle_add(gateway_suspend_complete, owner);
+	return id++;
+}
+
+static void cancel_gateway(struct media_transport *transport, guint id)
+{
+	gateway_cancel_stream(transport->device, id);
+}
+
 static void media_owner_exit(DBusConnection *connection, void *user_data)
 {
 	struct media_owner *owner = user_data;
@@ -489,6 +609,9 @@ static void media_transport_add(struct media_transport *transport,
 	DBG("Transport %s Owner %s", transport->path, owner->name);
 	transport->owners = g_slist_append(transport->owners, owner);
 	owner->transport = transport;
+	owner->watch = g_dbus_add_disconnect_watch(transport->conn, owner->name,
+							media_owner_exit,
+							owner, NULL);
 }
 
 static struct media_owner *media_owner_create(DBusConnection *conn,
@@ -500,9 +623,6 @@ static struct media_owner *media_owner_create(DBusConnection *conn,
 	owner = g_new0(struct media_owner, 1);
 	owner->name = g_strdup(dbus_message_get_sender(msg));
 	owner->accesstype = g_strdup(accesstype);
-	owner->watch = g_dbus_add_disconnect_watch(conn, owner->name,
-							media_owner_exit,
-							owner, NULL);
 
 	DBG("Owner created: sender=%s accesstype=%s", owner->name,
 			accesstype);
@@ -561,6 +681,7 @@ static DBusMessage *acquire(DBusConnection *conn, DBusMessage *msg,
 	owner = media_owner_create(conn, msg, accesstype);
 	id = transport->resume(transport, owner);
 	if (id == 0) {
+		media_transport_release(transport, accesstype);
 		media_owner_free(owner);
 		return btd_error_not_authorized(msg);
 	}
@@ -634,13 +755,30 @@ static int set_property_a2dp(struct media_transport *transport,
 						const char *property,
 						DBusMessageIter *value)
 {
+	struct a2dp_transport *a2dp = transport->data;
+
 	if (g_strcmp0(property, "Delay") == 0) {
 		if (dbus_message_iter_get_arg_type(value) != DBUS_TYPE_UINT16)
 			return -EINVAL;
-		dbus_message_iter_get_basic(value, &transport->delay);
+		dbus_message_iter_get_basic(value, &a2dp->delay);
 
 		/* FIXME: send new delay */
 		return 0;
+	} else if (g_strcmp0(property, "Volume") == 0) {
+		uint16_t volume;
+
+		if (dbus_message_iter_get_arg_type(value) != DBUS_TYPE_UINT16)
+			return -EINVAL;
+
+		dbus_message_iter_get_basic(value, &volume);
+
+		if (volume > 127)
+			return -EINVAL;
+
+		if (a2dp->volume == volume)
+			return 0;
+
+		return avrcp_set_volume(transport->device, volume);
 	}
 
 	return -EINVAL;
@@ -673,6 +811,13 @@ static int set_property_headset(struct media_transport *transport,
 	return -EINVAL;
 }
 
+static int set_property_gateway(struct media_transport *transport,
+						const char *property,
+						DBusMessageIter *value)
+{
+	return -EINVAL;
+}
+
 static DBusMessage *set_property(DBusConnection *conn, DBusMessage *msg,
 								void *data)
 {
@@ -722,7 +867,13 @@ static DBusMessage *set_property(DBusConnection *conn, DBusMessage *msg,
 static void get_properties_a2dp(struct media_transport *transport,
 						DBusMessageIter *dict)
 {
-	dict_append_entry(dict, "Delay", DBUS_TYPE_UINT16, &transport->delay);
+	struct a2dp_transport *a2dp = transport->data;
+
+	dict_append_entry(dict, "Delay", DBUS_TYPE_UINT16, &a2dp->delay);
+
+	if (a2dp->volume <= 127)
+		dict_append_entry(dict, "Volume", DBUS_TYPE_UINT16,
+							&a2dp->volume);
 }
 
 static void get_properties_headset(struct media_transport *transport,
@@ -741,6 +892,12 @@ static void get_properties_headset(struct media_transport *transport,
 	dict_append_entry(dict, "Routing", DBUS_TYPE_STRING, &routing);
 }
 
+static void get_properties_gateway(struct media_transport *transport,
+						DBusMessageIter *dict)
+{
+	/* None */
+}
+
 void transport_get_properties(struct media_transport *transport,
 							DBusMessageIter *iter)
 {
@@ -790,21 +947,50 @@ static DBusMessage *get_properties(DBusConnection *conn, DBusMessage *msg,
 	return reply;
 }
 
-static GDBusMethodTable transport_methods[] = {
-	{ "GetProperties",	"",	"a{sv}",	get_properties },
-	{ "Acquire",		"s",	"h",		acquire,
-						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "Release",		"s",	"",		release,
-						G_DBUS_METHOD_FLAG_ASYNC},
-	{ "SetProperty",	"sv",	"",		set_property },
+static const GDBusMethodTable transport_methods[] = {
+	{ GDBUS_METHOD("GetProperties",
+			NULL, GDBUS_ARGS({ "properties", "a{sv}" }),
+			get_properties) },
+	{ GDBUS_ASYNC_METHOD("Acquire",
+			GDBUS_ARGS({ "access_type", "s" }),
+			GDBUS_ARGS({ "fd", "h" }, { "mtu_r", "q" },
+							{ "mtu_w", "q" } ),
+			acquire) },
+	{ GDBUS_ASYNC_METHOD("Release",
+			GDBUS_ARGS({ "access_type", "s" }), NULL,
+			release ) },
+	{ GDBUS_ASYNC_METHOD("SetProperty",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" }),
+			NULL, set_property) },
 	{ },
 };
 
-static GDBusSignalTable transport_signals[] = {
-	{ "PropertyChanged",	"sv"	},
+static const GDBusSignalTable transport_signals[] = {
+	{ GDBUS_SIGNAL("PropertyChanged",
+			GDBUS_ARGS({ "name", "s" }, { "value", "v" })) },
 	{ }
 };
 
+static void destroy_a2dp(void *data)
+{
+	struct a2dp_transport *a2dp = data;
+
+	if (a2dp->session)
+		avdtp_unref(a2dp->session);
+
+	g_free(a2dp);
+}
+
+static void destroy_headset(void *data)
+{
+	struct headset_transport *headset = data;
+
+	if (headset->nrec_id > 0)
+		headset_remove_nrec_cb(headset->device, headset->nrec_id);
+
+	g_free(headset);
+}
+
 static void media_transport_free(void *data)
 {
 	struct media_transport *transport = data;
@@ -818,11 +1004,8 @@ static void media_transport_free(void *data)
 
 	g_slist_free(transport->owners);
 
-	if (transport->session)
-		avdtp_unref(transport->session);
-
-	if (transport->nrec_id)
-		headset_remove_nrec_cb(transport->device, transport->nrec_id);
+	if (transport->destroy != NULL)
+		transport->destroy(transport->data);
 
 	if (transport->conn)
 		dbus_connection_unref(transport->conn);
@@ -867,21 +1050,42 @@ struct media_transport *media_transport_create(DBusConnection *conn,
 	uuid = media_endpoint_get_uuid(endpoint);
 	if (strcasecmp(uuid, A2DP_SOURCE_UUID) == 0 ||
 			strcasecmp(uuid, A2DP_SINK_UUID) == 0) {
+		struct a2dp_transport *a2dp;
+
+		a2dp = g_new0(struct a2dp_transport, 1);
+		a2dp->volume = -1;
+
 		transport->resume = resume_a2dp;
 		transport->suspend = suspend_a2dp;
 		transport->cancel = cancel_a2dp;
 		transport->get_properties = get_properties_a2dp;
 		transport->set_property = set_property_a2dp;
+		transport->data = a2dp;
+		transport->destroy = destroy_a2dp;
 	} else if (strcasecmp(uuid, HFP_AG_UUID) == 0 ||
 			strcasecmp(uuid, HSP_AG_UUID) == 0) {
+		struct headset_transport *headset;
+
+		headset = g_new0(struct headset_transport, 1);
+		headset->device = device;
+		headset->nrec_id = headset_add_nrec_cb(device,
+							headset_nrec_changed,
+							transport);
+
 		transport->resume = resume_headset;
 		transport->suspend = suspend_headset;
 		transport->cancel = cancel_headset;
 		transport->get_properties = get_properties_headset;
 		transport->set_property = set_property_headset;
-		transport->nrec_id = headset_add_nrec_cb(device,
-							headset_nrec_changed,
-							transport);
+		transport->data = headset;
+		transport->destroy = destroy_headset;
+	} else if (strcasecmp(uuid, HFP_HS_UUID) == 0 ||
+			strcasecmp(uuid, HSP_HS_UUID) == 0) {
+		transport->resume = resume_gateway;
+		transport->suspend = suspend_gateway;
+		transport->cancel = cancel_gateway;
+		transport->get_properties = get_properties_gateway;
+		transport->set_property = set_property_gateway;
 	} else
 		goto fail;
 
@@ -908,13 +1112,36 @@ const char *media_transport_get_path(struct media_transport *transport)
 void media_transport_update_delay(struct media_transport *transport,
 							uint16_t delay)
 {
+	struct a2dp_transport *a2dp = transport->data;
+
 	/* Check if delay really changed */
-	if (transport->delay == delay)
+	if (a2dp->delay == delay)
 		return;
 
-	transport->delay = delay;
+	a2dp->delay = delay;
 
 	emit_property_changed(transport->conn, transport->path,
 				MEDIA_TRANSPORT_INTERFACE, "Delay",
-				DBUS_TYPE_UINT16, &transport->delay);
+				DBUS_TYPE_UINT16, &a2dp->delay);
+}
+
+struct audio_device *media_transport_get_dev(struct media_transport *transport)
+{
+	return transport->device;
+}
+
+void media_transport_update_volume(struct media_transport *transport,
+								uint8_t volume)
+{
+	struct a2dp_transport *a2dp = transport->data;
+
+	/* Check if volume really changed */
+	if (a2dp->volume == volume)
+		return;
+
+	a2dp->volume = volume;
+
+	emit_property_changed(transport->conn, transport->path,
+				MEDIA_TRANSPORT_INTERFACE, "Volume",
+				DBUS_TYPE_UINT16, &a2dp->volume);
 }
