@@ -237,16 +237,6 @@ static int request_encryption(bdaddr_t *src, bdaddr_t *dst)
 	return err;
 }
 
-static void enable_sixaxis(int csk)
-{
-	const unsigned char buf[] = {
-		0x53 /*HIDP_TRANS_SET_REPORT | HIDP_DATA_RTYPE_FEATURE*/,
-		0xf4,  0x42, 0x03, 0x00, 0x00 };
-	int err;
-
-	err = write(csk, buf, sizeof(buf));
-}
-
 static int create_device(int ctl, int csk, int isk, uint8_t subclass, int nosdp, int nocheck, int bootonly, int encrypt, int timeout)
 {
 	struct hidp_connadd_req req;
@@ -335,14 +325,10 @@ create:
 		req.flags |= (1 << HIDP_BOOT_PROTOCOL_MODE);
 	}
 
-	if (req.vendor == 0x054c && req.product == 0x0268)
-		enable_sixaxis(csk);
-
 	err = ioctl(ctl, HIDPCONNADD, &req);
 
 error:
-	if (req.rd_data)
-		free(req.rd_data);
+	free(req.rd_data);
 
 	return err;
 }
@@ -573,7 +559,7 @@ static void do_search(int ctl, bdaddr_t *bdaddr, uint8_t subclass, int fakehid,
 
 	for (i = 0; i < num_rsp; i++) {
 		memcpy(class, (info+i)->dev_class, 3);
-		if ((class[0] == 0x00 && class[2] == 0x00 && 
+		if ((class[0] == 0x00 && class[2] == 0x00 &&
 				(class[1] == 0x40 || class[1] == 0x1f)) ||
 				(class[0] == 0x10 && class[1] == 0x02 && class[2] == 0x40)) {
 			bacpy(&dst, &(info+i)->bdaddr);
@@ -825,7 +811,7 @@ int main(int argc, char *argv[])
         if (detach) {
 		if (daemon(0, 0)) {
 			perror("Can't start daemon");
-        	        exit(1);
+			exit(1);
 		}
 	} else
 		log_option |= LOG_PERROR;
