@@ -44,10 +44,11 @@
 #include <glib.h>
 #include <dbus/dbus.h>
 
+#include "hcid.h"
 #include "sdpd.h"
 #include "log.h"
-#include "manager.h"
 #include "adapter.h"
+#include "manager.h"
 
 static sdp_record_t *server = NULL;
 
@@ -65,7 +66,7 @@ static const int sdpServerVnumEntries = 1;
  * seconds. Used for updating the service db state
  * attribute of the service record of the SDP server
  */
-uint32_t sdp_get_time()
+uint32_t sdp_get_time(void)
 {
 	/*
 	 * To handle failure in gettimeofday, so an old
@@ -93,22 +94,6 @@ static void update_db_timestamp(void)
 	sdp_attr_replace(server, SDP_ATTR_SVCDB_STATE, d);
 }
 
-static void update_svclass_list(const bdaddr_t *src)
-{
-	GSList *adapters = manager_get_adapters();
-
-	for (; adapters; adapters = adapters->next) {
-		struct btd_adapter *adapter = adapters->data;
-		bdaddr_t bdaddr;
-
-		adapter_get_address(adapter, &bdaddr);
-
-		if (bacmp(src, BDADDR_ANY) == 0 || bacmp(src, &bdaddr) == 0)
-			btd_adapter_services_updated(adapter);
-	}
-
-}
-
 void register_public_browse_group(void)
 {
 	sdp_list_t *browselist;
@@ -186,13 +171,11 @@ void register_server_service(void)
 	sdp_attr_add(server, SDP_ATTR_VERSION_NUM_LIST, pData);
 
 	update_db_timestamp();
-	update_svclass_list(BDADDR_ANY);
 }
 
-void register_device_id(const uint16_t vendor, const uint16_t product,
-						const uint16_t version)
+void register_device_id(void)
 {
-	const uint16_t spec = 0x0102, source = 0x0002;
+	const uint16_t spec = 0x0103;
 	const uint8_t primary = 1;
 	sdp_list_t *class_list, *group_list, *profile_list;
 	uuid_t class_uuid, group_uuid;
@@ -201,9 +184,9 @@ void register_device_id(const uint16_t vendor, const uint16_t product,
 	sdp_profile_desc_t profile;
 	sdp_record_t *record = sdp_record_alloc();
 
-	info("Adding device id record for %04x:%04x", vendor, product);
-
-	btd_manager_set_did(vendor, product, version);
+	info("Adding device id record for %04x:%04x:%04x:%04x",
+				main_opts.did_source, main_opts.did_vendor,
+				main_opts.did_product, main_opts.did_version);
 
 	record->handle = sdp_next_handle();
 
@@ -230,23 +213,22 @@ void register_device_id(const uint16_t vendor, const uint16_t product,
 	spec_data = sdp_data_alloc(SDP_UINT16, &spec);
 	sdp_attr_add(record, 0x0200, spec_data);
 
-	vendor_data = sdp_data_alloc(SDP_UINT16, &vendor);
+	vendor_data = sdp_data_alloc(SDP_UINT16, &main_opts.did_vendor);
 	sdp_attr_add(record, 0x0201, vendor_data);
 
-	product_data = sdp_data_alloc(SDP_UINT16, &product);
+	product_data = sdp_data_alloc(SDP_UINT16, &main_opts.did_product);
 	sdp_attr_add(record, 0x0202, product_data);
 
-	version_data = sdp_data_alloc(SDP_UINT16, &version);
+	version_data = sdp_data_alloc(SDP_UINT16, &main_opts.did_version);
 	sdp_attr_add(record, 0x0203, version_data);
 
 	primary_data = sdp_data_alloc(SDP_BOOL, &primary);
 	sdp_attr_add(record, 0x0204, primary_data);
 
-	source_data = sdp_data_alloc(SDP_UINT16, &source);
+	source_data = sdp_data_alloc(SDP_UINT16, &main_opts.did_source);
 	sdp_attr_add(record, 0x0205, source_data);
 
 	update_db_timestamp();
-	update_svclass_list(BDADDR_ANY);
 }
 
 int add_record_to_server(const bdaddr_t *src, sdp_record_t *rec)
@@ -287,7 +269,6 @@ int add_record_to_server(const bdaddr_t *src, sdp_record_t *rec)
 	}
 
 	update_db_timestamp();
-	update_svclass_list(src);
 
 	return 0;
 }
@@ -296,16 +277,18 @@ int remove_record_from_server(uint32_t handle)
 {
 	sdp_record_t *rec;
 
+	/* Refuse to remove the server's own record */
+	if (handle == SDP_SERVER_RECORD_HANDLE)
+		return -EINVAL;
+
 	DBG("Removing record with handle 0x%05x", handle);
 
 	rec = sdp_record_find(handle);
 	if (!rec)
 		return -ENOENT;
 
-	if (sdp_record_remove(handle) == 0) {
+	if (sdp_record_remove(handle) == 0)
 		update_db_timestamp();
-		update_svclass_list(BDADDR_ANY);
-	}
 
 	sdp_record_free(rec);
 
@@ -432,7 +415,7 @@ int service_register_req(sdp_req_t *req, sdp_buf_t *rsp)
 		bufsize -= sizeof(bdaddr_t);
 	}
 
-	// save image of PDU: we need it when clients request this attribute
+	/* save image of PDU: we need it when clients request this attribute */
 	rec = extract_pdu_server(&req->device, p, bufsize, 0xffffffff, &scanned);
 	if (!rec)
 		goto invalid;
@@ -469,7 +452,6 @@ success:
 	}
 
 	update_db_timestamp();
-	update_svclass_list(&req->device);
 
 	/* Build a rsp buffer */
 	bt_put_unaligned(htonl(rec->handle), (uint32_t *) rsp->data);
@@ -518,7 +500,6 @@ int service_update_req(sdp_req_t *req, sdp_buf_t *rsp)
 	assert(nrec == orec);
 
 	update_db_timestamp();
-	update_svclass_list(BDADDR_ANY);
 
 done:
 	p = rsp->data;
@@ -538,17 +519,14 @@ int service_remove_req(sdp_req_t *req, sdp_buf_t *rsp)
 	int status = 0;
 
 	/* extract service record handle */
-	p += sizeof(uint32_t);
 
 	rec = sdp_record_find(handle);
 	if (rec) {
 		sdp_svcdb_collect(rec);
 		status = sdp_record_remove(handle);
 		sdp_record_free(rec);
-		if (status == 0) {
+		if (status == 0)
 			update_db_timestamp();
-			update_svclass_list(BDADDR_ANY);
-		}
 	} else {
 		status = SDP_INVALID_RECORD_HANDLE;
 		SDPDBG("Could not find record : 0x%x", handle);
